import sys
import os
import uuid
import io
import traceback
import json
import numpy as np
import pandas as pd
import dill
import inspect
import shutil
import base64
import tempfile

vals = {}

class NumpyAndPandasEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            if obj.ndim == 1:
                return obj.tolist()
            else:
                return [self.default(obj[i]) for i in range(obj.shape[0])]
        elif isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, pd.DataFrame):
            d = obj.to_dict(orient="list")
            dict_to_encode = {str(key): d[key] for key in d}
            if (type(obj.index) is not pd.RangeIndex):
                dict_to_encode["index"] = list(obj.index)
            return dict_to_encode
        return json.JSONEncoder.default(self, obj)

def load_environment(filename):
    global vals
    old_stdout = sys.stdout
    sys.stdout = io.StringIO()
    old_stderr = sys.stderr
    sys.stderr = io.StringIO()
    try:
        with open(filename,'rb') as file:
            vals = dill.load(file)
        return True
    except Exception as e:
        return str(e)
    finally:
        sys.stderr = old_stderr
        sys.stdout = old_stdout

def dump_environment(filename):
    global vals
    vals = remove_closed_files(vals)
    try:
        with open(filename,'wb') as file:
            dill.dump(vals, file)
        return True
    except Exception as e: 
        return str(e)

def remove_closed_files(vals: dict):
    new_vals = {}
    for key, value in vals.items():
        try:
           if value.closed: continue
           new_vals[key] = value
        except:
           new_vals[key] = value
    return new_vals

def set_object(filename,objectname):
    global vals
    old_stdout = sys.stdout
    sys.stdout = io.StringIO()
    old_stderr = sys.stderr
    sys.stderr = io.StringIO()
    try:
        with open(filename,'rb') as file:
            obj = dill.load(file)
        if inspect.isfunction(obj):
            exec('def '+objectname+str(inspect.signature(obj))+':\n pass\n',vals)
            vals[objectname].__code__ = obj.__code__
        else:
            vals[objectname] = obj
        return True
    except Exception as e:
        return str(e)
    finally:
        sys.stderr = old_stderr
        sys.stdout = old_stdout


def write_object_to_file(filename, objectname):
    global vals
    try:
        with open(filename,'wb') as file: 
            dill.dump(vals[objectname],file)
        return True
    except KeyError as e:
        return "Object name does not exist"
    except:
        return traceback.format_exc()

def delete_object(objectname):
    try:
        del vals[objectname]
        return True
    except KeyError as e:
        return "Object name does not exist"
    except:
        return traceback.format_exc()

def cleanup_and_exit():
    sys.stdout.close()
    sys.stderr.close()
    sys.exit(1)


def reset_environment():
    global vals
    vals = {}
    return True

def execution_result(success, stdout, stderr):
    return {"success": success, "console_output": stdout, "error_message": stderr}


def map_type(obj):
    if isinstance(obj, list):
        return np.array(obj)
    if isinstance(obj, dict):
        try:
            ret_dict = pd.DataFrame(obj)
            return ret_dict
        except:
            return obj
    return obj

JsonString = str

def get_workspace_objects():
    global vals
    return list(vals.keys())

def set_values_and_execute(code: list, input_parameters: JsonString, output_parameter_names:JsonString):
    global vals
    parsed_vals = json.loads(input_parameters)
    new_vals = {key: map_type(parsed_vals[key]) for key in parsed_vals}
    vals.update(new_vals)
    params = json.loads(output_parameter_names)
    ret = do_execution(os.linesep.join(code), vals, None)
    ret["output_parameters"] = {key: vals[key] for key in params if key in vals}
    try:
        ret_str = json.dumps(ret, cls=NumpyAndPandasEncoder)
    except TypeError as e:
        ret_str = json.dumps(execution_result(False,"","Error Encoding Value {0}".format(e)))
    finally:
        print(ret_str)


def do_execution(code_string, env, local_values):
    old_stdout = sys.stdout
    sys.stdout = io.StringIO()
    old_stderr = sys.stderr
    sys.stderr = io.StringIO()
    result = None
    exception = None
    try:
        code = compile(code_string, "<REMOTE>", "exec")
        eval(code, env, local_values)
        success = True
        exception = ""
    except Exception as e:
        success = False
        exception = traceback.format_exception(type(e),e,e.__traceback__)
        exception = "".join([v for i, v in enumerate(exception) if i not in [1,2]])
    finally:
        result = execution_result(success, sys.stdout.getvalue(), exception)
        sys.stderr = old_stderr
        sys.stdout = old_stdout
        return result

